


#########################################################################################################
#                                                                                                       #
#                                        PREDICTION MODEL                                               #
#                                        ================                                               #
#                                                                                                       #
# Luisa Mennicken                                                                                       #
# Master thesis                                                                                         #
#                                                                                                       #
#########################################################################################################


### Necessary libraries

library(lubridate)
library(klaR)
library(MASS)
library(stats)
library(caTools)
library(FactoMineR)
library(caret)
library(ROCR)
library(MVN)
library(Hotelling)
library(nortest)
library(ggplot2)
library(ggpubr)
library(viridis)
theme_set(theme_pubr())


### Lecture and storage of the data

data=read.csv2("C:/Users/luisa/UNI/Memoire/2017_10_30_Pretraitements/FinalDataWithoutBaselineLogAna.csv")
head(data)
dim(data)

### Parsing into date-time object

data$time=parse_date_time(data$time,orders="ymd HMS")
time=data$time


#########################################################################################################
#                                                                                                       #
#                                   PREDICTION OVER COMPLETE DATA                                       #
#                                   =============================                                       #
#                                                                                                       #
# Input: - data: data set                                                                               #
#        - anaIndex: integer between 1 and 6 (choice of analyser)                                       #
#        - lim: real number defining the threshold for the binary variable                              #
#                                                                                                       #
# Output: The function "predictionsOverAllDataPlots" performs a LDA model with selection of variables,  #
#         the MLR model with these selected variables and return several results from the two models:   #
#         the predictions on the real analysers measurements, the evolution of the accuracy and         #
#         adjusted R squares over the selection, the confusion matrix, the ROC curve, the plot of true  #
#         and false predicted signals and the contribution plot.                                        #
#                                                                                                       #
#########################################################################################################


predictionsOverAllDataPlots<-function(data,anaIndex,lim){
  
  analysers = data[,c("A_CH4_ppm" , 
                      "A_H2S_g.m3",
                      "A_NH3_g.m3",
                      "A_BENZ_g.m3" ,
                      "A_TOLU_g.m3",
                      "A_LIMO_g.m3")]
  
  sensors=data[,c("S_2602_kOhm",
                   "S_2610_kOhm", 
                   "S_2611_kOhm", 
                   "S_2620_kOhm", 
                   "S_1330_kOhm", 
                   "S_2444_kOhm")]
  
  ### Cration des variables correspondant aux interactions multiples
  
  interactions= model.matrix(~.^6, data=sensors)[,-1]
  print(dim(interactions))
  
  workData=cbind2(data[,c(1:7)],interactions)
  print(head(workData))
  
  ### select only the specified analyser and all interactions
  
  dataPCA=workData[,c(anaIndex+1,8:ncol(workData))]
  print(head(dataPCA))
  
  ### PCA over the interactions of sensors
  
  res.pca = PCA(dataPCA, scale.unit = TRUE, ncp = ncol(dataPCA), graph = TRUE, 
                quanti.sup = 1)
  
  Variables_PCA_Sensors = res.pca$var$coord # coordinates of variables 
  PCA_Sensors_trainSet = res.pca$ind$coord #projection of measurements on decorrelated variables
  
  ### Creation of binary analyser variable
  
  ana=dataPCA[,1]
  plot(data$time,exp(ana),type="l")
  abline(h=exp(lim),col="red")
  signal=ifelse(ana>=lim,"signal","noise")
  signal_indices<-which(signal=="signal")
  noise_indices<-which(signal=="noise")
  
  ### Stepwise selection of the LDA model
  
  model=stepclass(x=PCA_Sensors_trainSet,method="lda",grouping=signal,improvement=0.001,criterion="AC",direction="forward")
  
  m=model$process[-1,]
  
  barplot(m[,4]*10,names.arg=m[,3],ylab="Accuracy [%]",
          main="Accuracy in terms of selected variables",las=2,ylim=c(0,100))
  
  cat("------ percentage of variance \n")
  print(res.pca$eig[m[,2],2])
  
  ### LDA model with selected variables
  
  newModel=lda(x=as.matrix(PCA_Sensors_trainSet[,model$model[,1]]),grouping=signal)
  
  newModel.pred=predict(object=newModel, type="response")
  
  mainText=c("CH4","H2S","NH3","BENZ","TOLU","LIMO")
  analysersUnits=c("[ppm]",rep("[g/m3]",5))
  
  plot(time,exp(ana),xlab="time [30 min]",ylab=paste(mainText[anaIndex],analysersUnits[anaIndex]),
       main=paste("LDA model -", mainText[anaIndex] ,"analyser"),type="l")
  abline(h=exp(lim),col="red")
  axis(2,at=exp(lim))
  lines(time,ifelse(newModel.pred$class=="signal",400,0),type="h",col=" light green")
  lines(time,exp(ana))
  
  ### Comparison TrueSignal and PredictedSignal
  
  TrueSignal=ifelse(signal=="signal",1,0)
  PredictedSignal=ifelse(newModel.pred$class=="signal",1,0)
  plot(workData$time,TrueSignal-PredictedSignal,type="l",xlab="time [30 min]",
       main="Difference of true and predicted signals on the whole data set")
  
  
  plot(time,exp(ana),xlab="time [30 min]",ylab=paste(mainText[anaIndex],analysersUnits[anaIndex]),
       main=paste("Prediction of", mainText[anaIndex] ,"signal with LDA"),pch=20)
  FN=time[which(TrueSignal==1 & PredictedSignal==0)]
  FP=time[which(TrueSignal==0 & PredictedSignal==1)]
  
  points(FN,exp(ana)[which(TrueSignal==1 & PredictedSignal==0)],col="red",pch=20)
  points(FP,exp(ana)[which(TrueSignal==0 & PredictedSignal==1)],col="green",pch=20)
  abline(h=exp(lim),col="orange")
  legend("topleft",legend=c("False negative","False positive","Right positive and negatives"),
         col=c("red","green","black"),pch=20)
  
  ### ConfusionMatrix
  
  print(confusionMatrix(reference=signal,data=newModel.pred$class,positive="signal"))
  
  ### ROC Curve
  
  pred<-prediction(predictions=newModel.pred$posterior[,2],labels=signal)
  perf<-performance(pred,"tpr","fpr")
  plot(perf,main=paste("ROC Curve -",mainText[anaIndex],"analyser"),xlab="False positive rate",
       ylab="True positive rate")
  
  ### Multiple regression
  
  TrainingSet=cbind2(dataPCA,PCA_Sensors_trainSet[,model$model[,1]])
  PCAVarNames= colnames(PCA_Sensors_trainSet)[model$model[,1]]
  formula=reformulate(termlabels = c(PCAVarNames), response = colnames(analysers)[anaIndex])
  modStep=lm(formula,data=TrainingSet)
  print(summary(modStep))
  
  
  adjR=NULL
  for(i in 1:length(m[,3])){
    modInter=lm(data[,anaIndex+1]~.,data=as.data.frame(TrainingSet[,as.character(m[1:i,3])]))
    adjR=c(adjR,summary(modInter)$adj.r.squared)
    
  }
  cat("Adjusted R squares\n")
  print(adjR)
  
  barplot(adjR,names.arg=m[,3],ylab="Adjusted R squared",
          main="Adjusted R squared in terms of \n selected variables",las=2,ylim=c(0,1))
  
  estimation = modStep$fitted.values
  predictors =  names(coef(modStep))[-1]
  print(predictors)
  prediction = predict(modStep, type="response")
  
  plot(time,exp(ana),type="l",xlab="time [30 min]",ylab=paste(mainText[anaIndex],analysersUnits[anaIndex]),
       main=paste("MLR model -",mainText[anaIndex],"analyser"))
  lines(time,exp(modStep$fitted.values),col="light green",type="h")
  lines(time,exp(ana),type="l",lty=3)
  
  plot(time,exp(ana),type="l",xlab="time [30 min]",ylab=paste(mainText[anaIndex],analysersUnits[anaIndex]),
       main=paste("MLR model -",mainText[anaIndex],"analyser"))
  lines(time,ifelse(newModel.pred$class=="signal",400,0),type="h",col=" orange")
  lines(time,exp(modStep$fitted.values),col="light green",type="h")
  lines(time,exp(ana),type="l")
  abline(h=exp(lim))
  
  ### Contributions
  
  contribLDA=res.pca$var$contrib[,model$model[,1]]
  print(contribLDA)
  print(summary(contribLDA))
  print(ggballoonplot(as.data.frame(contribLDA),fill="value",ggtheme=theme_bw())+scale_fill_viridis(option = "B"))
}


#### Application of the function predictionsOverAllDataPlots

predictionsOverAllDataPlots(data,1,log(5))



#########################################################################################################
#                                                                                                       #
#                                 Cross-validation with random split                                    #
#                                 ==================================                                    #
#                                                                                                       #
# Input: - data: data set                                                                               #
#        - anaIndex: integer between 1 and 6 (choice of analyser)                                       #
#        - lim: real number defining the threshold for the binary variable                              #
#                                                                                                       #
# Output: The function "selectedPC_randomSplit" performs random split over the data (80%-20%),          #
#         then performs the LDA model with selection on the training set, prints the selected           #
#         components and the confusion matrix over the test set, performs a MLR model on the selected   #
#         variables,applies it on the test set and returns a list of results:                           #
#         sensitivity, specificity, the coordinates for the ROC curve and the adjusted R squared.       #
#                                                                                                       #
#########################################################################################################

selectedPC_randomSplit<-function(data,anaIndex,lim){
  
  analysers = data[,c("A_CH4_ppm" , 
                      "A_H2S_g.m3",
                      "A_NH3_g.m3",
                      "A_BENZ_g.m3" ,
                      "A_TOLU_g.m3",
                      "A_LIMO_g.m3")]
  
  sensors=data[,c("S_2602_kOhm",
                  "S_2610_kOhm", 
                  "S_2611_kOhm", 
                  "S_2620_kOhm", 
                  "S_1330_kOhm", 
                  "S_2444_kOhm")]
  
  
  interactions= model.matrix(~.^6, data=sensors)[,-1]
  workData=cbind2(data[,c(1:7)],interactions) #time, analyser, interactions
  
  ### select only the specified analyser and all interactions
  
  dataPCA=workData[,c(anaIndex+1,8:ncol(workData))] #+1 because of the first column "time"
  ana=dataPCA[,1]
  
  ### Random split plit in training and test set 
  
  splitVector=sample.split(c(1:nrow(dataPCA)),SplitRatio=0.8)
  trainSetIndices=which(splitVector==TRUE)
  testSetIndices=which(splitVector==FALSE)
  
  ### ACP sur les interactions des capteurs
  
  res.pca = PCA(dataPCA, scale.unit = TRUE, ncp = ncol(dataPCA), graph = FALSE, 
                quanti.sup = 1,
                ind.sup = testSetIndices 
  )
  
  Variables_PCA_Sensors = res.pca$var$coord 
  PCA_Sensors_trainSet = res.pca$ind$coord 
  PCA_Sensors_testSet = res.pca$ind.sup$coord 
  
  ### Creation of binary analyser variable
  
  signal=ifelse(ana>=lim,"signal","noise")
  signal_train=signal[trainSetIndices]
  signal_test=signal[testSetIndices]
  
  ### LDA Model 
  model=stepclass(x=PCA_Sensors_trainSet,grouping=signal_train,method="lda",
                  improvement=0.001,criterion="AC",direction="forward")
  newModel=lda(x=as.matrix(PCA_Sensors_trainSet[,model$model[,1]]),grouping=signal_train)
  newModel.pred=predict(object=newModel,newdata=as.matrix(PCA_Sensors_testSet[,model$model[,1]]),type="response")
  newModel.pred.train=predict(object=newModel,type="response")
  
  m=model$process[-1,]
  cat('Selected principal components:\n')
  print(m[,3])
  cat('\n')
  
  tab=confusionMatrix(reference=signal_test,data=newModel.pred$class,positive="signal")
  cat('Confusion matrix:\n')
  print(tab)
  cat('\n')
  
  ### ROC Curve
  pred<-prediction(predictions=newModel.pred$posterior[,2],labels=signal_test)
  perf<-performance(pred,"tpr","fpr")
  
  
  ### Multiple regression
  
  TrainingSet=cbind2(dataPCA[trainSetIndices,],PCA_Sensors_trainSet[,m[,2]])
  ValidationSet= cbind2(dataPCA[testSetIndices,], PCA_Sensors_testSet[,m[,2]])
  PCAVarNames= colnames(PCA_Sensors_trainSet)[m[,2]]
  formula=reformulate(termlabels = c(PCAVarNames), response = colnames(analysers)[anaIndex])
  modStep=lm(formula,data=TrainingSet)
  print(summary(modStep))

  estimation = modStep$fitted.values
  predictors =  names(coef(modStep))[-1]
  prediction = predict(modStep, newdata= ValidationSet[, predictors], type="response" )
  
  return(list("sensitivity"=tab$byClass[[1]],"specificity"=tab$byClass[[2]],"xperf"=perf@x.values,"yperf"=perf@y.values,
              "rSquare"=summary(modStep)$adj.r.squared))
  
}


### Application of the function "selectedPC_randomSplit" twenty times

sensitivities=NULL
specificities=NULL
rSquares=NULL

for(i in 1:20){
  cat("---------------------------------------------------\n")
  cat("Cross-validation : time",i,"\n")
  s=selectedPC_randomSplit(data,1,log(5))
  print(class(unlist(s$xperf)))
  cat("---------------------------------------------------\n")
  
  sensitivities=c(sensitivities,s$sensitivity)
  specificities=c(specificities,s$specificity)
  if(i==1) plot(unlist(s$xperf),unlist(s$yperf),type="l",
                xlab="False positive rate",ylab="True positive rate",main="ROC curves - Cross validation")
  else lines(unlist(s$xperf),unlist(s$yperf),col=i)
  rSquares=c(rSquares,s$rSquare)
}

### Summaries of the results 

summary(sensitivities)
summary(specificities)
summary(rSquares)

### Sorted results 

sort(sensitivities)
sort(specificities)
sort(rSquares)



#########################################################################################################
#                                                                                                       #
#                               Prediction with split over time series                                  #
#                               ======================================                                  #
#                                                                                                       #
# Input: - data: data set                                                                               #
#        - anaIndex: integer between 1 and 6 (choice of analyser)                                       #
#        - lim: real number defining the threshold for the binary variable                              #
#                                                                                                       #
# Output: The function "timeSeriesSplitLDAandMLRPredPlot" performs a split on the sequence of time in   #
#         80% - 20%, then performs the LDA model with selection and the MLR model with the selected     #
#         variables, and prints the confusion matrix over the test set. A plot containing the           #
#         predictions of the LDA and MLR model for training and test set and the real measurements of   #
#         tha analyser is returned.                                                                     #
#                                                                                                       #
#########################################################################################################


timeSeriesSplitLDAandMLRPredPlot<-function(data,anaIndex,lim){
  
  analysers = data[,c("A_CH4_ppm" , 
                      "A_H2S_g.m3",
                      "A_NH3_g.m3",
                      "A_BENZ_g.m3" ,
                      "A_TOLU_g.m3",
                      "A_LIMO_g.m3")]
  
  sensors=data[,c("S_2602_kOhm",
                  "S_2610_kOhm", 
                  "S_2611_kOhm", 
                  "S_2620_kOhm", 
                  "S_1330_kOhm", 
                  "S_2444_kOhm")]
  
  interactions= model.matrix(~.^6, data=sensors)[,-1]
  workData=cbind2(data[,c(1:7)],interactions)
  
  
  ### select only the specified analyser and all interactions
  
  dataPCA=workData[,c(anaIndex+1,8:ncol(workData))]
  
  ### Split in training and test set with time series
  
  n=ceiling(80/100*nrow(dataPCA))
  trainSetIndices=1:n
  testSetIndices=(n+1):nrow(dataPCA)
  
  ### PCA on sensors interactions
  
  res.pca = PCA(dataPCA, scale.unit = TRUE, ncp = ncol(dataPCA), graph = TRUE, 
                quanti.sup = 1,
                ind.sup = testSetIndices # observations in validation set
  )
  
  Variables_PCA_Sensors = res.pca$var$coord 
  PCA_Sensors_trainSet = res.pca$ind$coord 
  PCA_Sensors_testSet = res.pca$ind.sup$coord 
  
  ### Creation of binary analyser variable
  
  ana=dataPCA[,1]
  signal=ifelse(ana>=lim,"signal","noise")
  signal_indices<-which(signal=="signal")
  noise_indices<-which(signal=="noise")
  
  signal_train=signal[trainSetIndices]
  signal_test=signal[testSetIndices]
  
  ### LDA Model 
  
  model=stepclass(x=PCA_Sensors_trainSet,grouping=signal_train,method="lda",improvement=0.001,criterion="AC",direction="forward")
  newModel=lda(x=as.matrix(PCA_Sensors_trainSet[,model$model[,1]]),grouping=signal_train)
  newModel.pred=predict(object=newModel,newdata=as.matrix(PCA_Sensors_testSet[,model$model[,1]]),
                        type="response")
  newModel.pred.train=predict(object=newModel, type="response")
  
  print(confusionMatrix(reference=signal_test,data=newModel.pred$class,positive="signal"))
  m=model$process[-1,]
  
  ### Multiple regression
  
  TrainingSet=cbind2(dataPCA[trainSetIndices,],PCA_Sensors_trainSet[,m[,2]])
  ValidationSet= cbind2(dataPCA[testSetIndices,], PCA_Sensors_testSet[,m[,2]])
  PCAVarNames= colnames(PCA_Sensors_trainSet)[m[,2]]
  formula=reformulate(termlabels = c(PCAVarNames), response = colnames(analysers)[anaIndex])
  modStep=lm(formula,data=TrainingSet)
  print(summary(modStep))
  
  estimation = modStep$fitted.values
  predictors =  names(coef(modStep))[-1]
  prediction = predict(modStep, newdata= ValidationSet[, predictors], type="response" )
  
  
  ### Plot
  
  mainText=c("CH4","H2S","NH3","BENZ","TOLU","LIMO")
  analysersUnits=c("[ppm]",rep("[g/m3]",5))
  par(mfrow=c(1,1))
  
  plot(time,exp(ana),type="l",xlab="time [30 min]",ylab=paste(mainText[anaIndex],analysersUnits[anaIndex]),
       main=paste( mainText[anaIndex],"predictions"))
  lines(time[testSetIndices],ifelse(newModel.pred$class=="signal",400,0),type="h",col=" light green")
  lines(time[trainSetIndices],ifelse(newModel.pred.train$class=="signal",400,0),type="h",col="light blue")
  lines(time[testSetIndices],exp(prediction),type="h",col="orange")
  lines(time[trainSetIndices],exp(estimation),type="h",col="indianred2")
  lines(time,exp(ana),type="l")
  abline(h=exp(lim))
  axis(side=2,at=exp(lim))
  legend("topright",legend=c(paste(mainText[anaIndex],"concentration"),"LDA on train set","LDA on test set","MLR on train set","MLR on test set"),
         col=c("black","light blue","light green","indianred2","orange"),lty=1,cex=0.8)
  
}


### Application of the function timeSeriesSplitLDAandMLRPredPlot

timeSeriesSplitLDAandMLRPredPlot(data,1,log(5))



#########################################################################################################
#                                                                                                       #
#                   Residuals in terms of response variable and control parameters                      #
#                   ==============================================================                      #                                   #
#                                                                                                       #
#########################################################################################################

### Lecture and Storage of the control parameters

controlParam=read.csv2("C:/Users/luisa/UNI/Memoire/2017_10_30_Pretraitements/controlParametersDataWithDeletedObs.csv")
head(controlParam)
dim(controlParam)

controlParam=na.omit(controlParam)


### Parsing into date-time object

controlParam$time=parse_date_time(controlParam$time,orders="ymd HMS")
controlTime=controlParam$time

analysers = data[,c("A_CH4_ppm" , 
                    "A_H2S_g.m3",
                    "A_NH3_g.m3",
                    "A_BENZ_g.m3" ,
                    "A_TOLU_g.m3",
                    "A_LIMO_g.m3")]

sensors=data[,c("S_2602_kOhm",
                "S_2610_kOhm", 
                "S_2611_kOhm", 
                "S_2620_kOhm", 
                "S_1330_kOhm", 
                "S_2444_kOhm")]

interactions= model.matrix(~.^6, data=sensors)[,-1]
dim(interactions)

workData=cbind2(data[,c(1:7)],interactions)
head(workData)

anaIndex=1
dataPCA=workData[,c(anaIndex+1,8:ncol(workData))]
head(dataPCA)

### PCA over the interactions of sensors

res.pca = PCA(dataPCA, scale.unit = TRUE, ncp = ncol(dataPCA), graph = TRUE, 
              quanti.sup = 1)
Variables_PCA_Sensors = res.pca$var$coord 
PCA_Sensors_trainSet = res.pca$ind$coord 


### Creation of binary analyser variable

ana=dataPCA[,1]

plot(data$time,exp(ana),type="l")
abline(h=5,col="red")
lim=log(5)
signal=ifelse(ana>=lim,"signal","noise")
signal_indices<-which(signal=="signal")
noise_indices<-which(signal=="noise")
length(signal_indices)
length(noise_indices)

### LDA model with stepwise selection

model=stepclass(x=PCA_Sensors_trainSet,method="lda",grouping=signal,improvement=0.001,criterion="AC",direction="forward")
newModel=lda(x=as.matrix(PCA_Sensors_trainSet[,model$model[,1]]),grouping=signal)
newModel.pred=predict(object=newModel, type="response")
TrueSignal=ifelse(signal=="signal",1,0)
PredictedSignal=ifelse(newModel.pred$class=="signal",1,0)

### MLR model

TrainingSet=cbind2(dataPCA,PCA_Sensors_trainSet[,model$model[,1]])
PCAVarNames= colnames(PCA_Sensors_trainSet)[model$model[,1]]
formula=reformulate(termlabels = c(PCAVarNames), response = colnames(analysers)[anaIndex])
modStep=lm(formula,data=TrainingSet)

### Plot of residuals in terms of the response variable

mainText=c("CH4","H2S","NH3","BENZ","TOLU","LIMO")
analysersUnits=c("[ppm]",rep("[g/m3]",5))
plot(ana,modStep$residuals,pch=20,xlab=paste("log values of",mainText[anaIndex],"analyser"),ylab="Residuals")
FN=time[which(TrueSignal==1 & PredictedSignal==0)]
FP=time[which(TrueSignal==0 & PredictedSignal==1)]

points(ana[which(TrueSignal==1 & PredictedSignal==0)],modStep$residuals[which(TrueSignal==1 & PredictedSignal==0)],col="red",pch=20)
points(ana[which(TrueSignal==0 & PredictedSignal==1)],modStep$residuals[which(TrueSignal==0 & PredictedSignal==1)],col="green",pch=20)
abline(v=lim)
axis(1,at=lim,label=FALSE)

### Plot of residuals in terms of the control parameters

par(mfrow=c(3,3))
for(i in 2:ncol(controlParam)){
  plot(controlParam[,i][1:5979],modStep$residuals,xlab=colnames(controlParam)[i],ylab="Residuals",pch=20)
  points(controlParam[which(TrueSignal==1 & PredictedSignal==0),i],modStep$residuals[which(TrueSignal==1 & PredictedSignal==0)],col="red",pch=20)
  points(controlParam[which(TrueSignal==0 & PredictedSignal==1),i],modStep$residuals[which(TrueSignal==0 & PredictedSignal==1)],col="green",pch=20)
  
}
par(mfrow=c(1,1))

