#########################################################################################################
#                                                                                                       #
#                                         PREPROCESSING                                                 #
#                                        ==============                                                 #
#                                                                                                       #
# Luisa Mennicken                                                                                       #
# Master thesis                                                                                         #
#                                                                                                       #
#########################################################################################################

### Necessary libraries

library(lubridate)
library(stats)

#########################################################################################################
#                                                                                                       #
#                                  1. Concatenation of sensors files                                    #
#                                  =================================                                    #
#                                                                                                       #
# Input: - files_location: quoted character string containing the path of the file where all the        #
#                          sensors data is stored                                                       #
#        - save_location: quoted character string containing the directory where the concatenation      #
#                         of the files will be saved                                                    #
#        - date: quoted string containing the column with date values                                   #
#                                                                                                       #
# Output: completeData: data set which is the concatenation of all the sensors files                    #
#                                                                                                       #
# Use: The function "sensorsFusion" concatenates all the files stored in "files_location" to one data   #
#      data set "completeData".                                                                         #
#                                                                                                       #
# Note: The function compares the dates in each file and checks that there are no repeated days.        #
#       If so, a warning message appears and the concerned file is not added to "completeData".         #
#                                                                                                       #
#########################################################################################################


sensorsFusion<-function(files_location,save_location,date){
  
  ### Create an empty data set. If it existed before, it will be emptied.
  
  file.create(save_location)
  files<-list.files(files_location)
  
  for(i in 1:length(files)){
    mon_doc<-paste0(paste0(files_location,"/"),files[i])
    f<-read.table(mon_doc,header=TRUE)
    
    if(i==1){
      date_1<-(read.table(file=mon_doc,header=TRUE,nrows=1)[,date])
      d1<-parse_date_time(date_1,orders="dmy")
      write.table(f,save_location,append=TRUE,col.names=TRUE,row.names=FALSE,dec=",",sep=";")
    }
    
    else{
      date_i<-(read.table(file=mon_doc,header=TRUE,nrows=1)[,date])
      di<-parse_date_time(date_i,orders="dmy")
      
      ### Check if the current date has been already seen before
      
      if(di==d1){
        cat(files[i])
        warning("Repeated day!")
      }
      
      else{
        d1<-di
        write.table(f,save_location,append=TRUE,col.names=FALSE,row.names=FALSE,dec=",",sep=";")
      }
    }
  }
  
  completeData=read.csv2(save_location,header=TRUE)
  return(completeData)
}



#########################################################################################################
#                                                                                                       #
#                                   2. Elimination of useless data                                      #
#                                   ==============================                                      #
#                                                                                                       #
# Input: - data: data set containing variables we want to retrieve                                      #
#        - var_to_keep: vector of character strings containing the desired column names of the data set #
#                       "data" we want to retain.                                                       #
#                                                                                                       #
# Output: newData: data set containing only the variables of "data" which are specified in "var_to_keep"#
#                                                                                                       #
# Use: The function reads the input data set "data" and creates a new data set containing only those    #
#       columns specified in the argument "var_to_keep".                                                #
#                                                                                                       #
# Note: In case of misspelling, a warning message appears and the data remains unchanged.               #
#         If we want to keep all variables, enter -1.                                                   #
#                                                                                                       #
#########################################################################################################

elimination<-function(data,var_to_keep){
  
  ### Check if the names are correct column names of the data
  notPossible<-which(!(var_to_keep %in% names(data)))
  
  if(length(notPossible)>0){
    warning("The specified columns to keep are not in the data set!")
    return(data)
  }
  
  else{
    newData<-data[ , var_to_keep]
  }
  return(newData)
}



#########################################################################################################
#                                                                                                       #
#                                   3. Parsing into date-time object                                    #
#                                   ================================                                    #
#                                                                                                       #
# Input: - data: data set containing a column with date values, and one with time values                #
#        - time_zone: character string indicating the time zone (UTC, MEZ,...)                          #
#        - date_column: quoted character string containing the column name of the date values           #
#        - time_column: quoted character string containing the column name of the time values           #  
#                                                                                                       #
# Output: newData: data set having date-time objects in the first column                                #
#                                                                                                       #
# Use: The function "createTimeSensors" reads the input data set, takes the date and time values,       #
#      parses them into date-time objects and returns a new data set with a first column containing the #
#      date-time objects, in the remaining columns the unchanged columns of "data".                     #
#                                                                                                       #
#########################################################################################################

createTimeSensors<-function(data,time_zone,date_column,time_column){
  
  time<-parse_date_time(paste(data[,date_column],data[,time_column]),orders="dmy HMS",tz=time_zone)
  del=which(colnames(data)==date_column | colnames(data)==time_column)
  newData<-data.frame(time,data[,-del])
  return(newData)
}


#########################################################################################################
#                                                                                                       #
#                                    4. Conductance transformation                                      #
#                                    =============================                                      #
#                                                                                                       #
# Input: - data: data set containing values to inverse                                                  #
#        - columns: vector containing the column names in quotation marks of the data set which have to #
#                   be inverted                                                                         #
#                                                                                                       #
# Output: newData: data set containing the inverse values of the elements in the specifiec columns      #
#                                                                                                       #
# Use: The function "conductance" transforms a data set by computing the inverse of the elements in the #
#      specified columns in the argument "columns".                                                     #
#                                                                                                       #
# Remark: In case of misspelling, the returned data set is the same as the input data set.              #
#                                                                                                       #
#########################################################################################################

conductance<-function(data,columns){
  
  notPossible<-which(!(columns %in% names(data)))
  
  if(length(notPossible)>0){
    warning("The specified columns to inverse are not possible!")
    newData=data
  }
  else{
    inverse<-function(x){
      return(1/x)
    }
    
    newData=data
    newData[,columns]=sapply(newData[,columns],inverse)
  }
  return(newData)
}


#########################################################################################################
#                                                                                                       #
#                                         5. Missing values                                             #
#                                         =================                                             #
#                                                                                                       #
# Input: - data: data frame containing in the first column date-time objects                            #
#                                                                                                       #
# Output: newData: input data set where rows have been added when a sensor measurement is missing       #
#                                                                                                       #
# Use: The function "addRowsNA" calculates the time differences between each consecutive couple of rows #
#      in "data" and if necessary inserts a row filled with NA and the corresponding time value.        #
#                                                                                                       #
#########################################################################################################

addRowsNA<-function(data){

  data$Dt = c(diff(data[,1]),0) # return the time difference in seconds - last one = 0 because no following value
  units(data$Dt) <-"mins"
  data$Dt=round(data$Dt)
  
  
  if (sum(data$Dt >1) >0){  # the following code is applied if we have at least one missing value
    
    lastMesureBeforeMissing = data$time[data$Dt >1] 
    firstMesureAfterMissing = data$time[which(data$Dt>1)+1]
    
    ###  Definition of the missing intervals 
    IntervalsMissing = cbind.data.frame(round(lastMesureBeforeMissing + minutes(1),units="mins"), 
                                        round(firstMesureAfterMissing - minutes(1),units="mins")) 
    
    colnames(IntervalsMissing) = c("Start", "End")

    cat('Intervals where the values are missing \n')
    cat('-------------------------------------- \n')
    print(IntervalsMissing)
    cat('-------------------------------------- \n')
    
    ### Preparation of the NA lignes to add
    
    ### Time list
    MissingDataTime = NULL
    for (i in 1:nrow(IntervalsMissing)){
      cat("i=",i,'\n')
      
      MissingDataTime= as_datetime(c(MissingDataTime, 
                                     seq(IntervalsMissing$Start[i], IntervalsMissing$End[i], by ="min")))
    }
    
    cat('Number of added lines in the file \n')
    cat('--------------------------------- \n')
    cat('N =', length(MissingDataTime), '\n')
    cat('Nb Obs file input :', nrow(data), '\n')
    cat('Expected Nb Obs file output :', nrow(data) + length(MissingDataTime), '\n')
    cat('--------------------------------- \n')
    
    ### Creation of data.frame with intercalated missings
    dataSupp = data.frame(MissingDataTime)
    names(dataSupp) = 'time'
    newData = merge(x= data, y= dataSupp, by= 'time', all.x= T, all.y = T)
    
    cat('Nb Obs file output :', nrow(newData), '\n')
    cat('--------------------------------- \n')
    
    
  }    
  else{newData = data}
  
  newData<-newData[,-ncol(newData)]
  return(newData)
}


#########################################################################################################
#                                                                                                       #
#                                 6. Read and save the analysers file                                   #
#                                 ===================================                                   #
#                                                                                                       #
# Input: - file_path: quoted string containing the file path of the analysers data set                  #
#                                                                                                       #
# Output: newData: data set containing the concatenation of the two first rows of the input data set as #
#         first row, and then the remaining values of this data set                                     #
#                                                                                                       #
# Use: The function "readAndSave" reads the input data set whose file path is "file_path", concatenates #
#      the two first lines of this data set and stores the output data set "newData".                   #
#                                                                                                       #
#########################################################################################################

readAndSave<-function(file_path){ 
  
  headerAna1 <- scan(file_path, nlines = 1,sep=";",what = character())
  headerAna2<-scan(file_path, nlines = 1,sep=";",what = character(),skip=1)
  pAna<-paste0(headerAna1,headerAna2)
  
  data_ana<- read.csv2(file_path, 
                       skip = 2,col.names=pAna,header = FALSE,na.strings="",sep=";",dec=",")
  
  return(data_ana)
}


#########################################################################################################
#                                                                                                       #
#                                 7. Elimination of useless data                                        #
#                                 ==============================                                        #
#                                                                                                       #
# We will use exactly the same function as for the sensors data (see line 79-112).                      #
#                                                                                                       #
#########################################################################################################


#########################################################################################################
#                                                                                                       #
#                                   8. Creation of date-time object                                     #
#                                   ===============================                                     #
#                                                                                                       #
# Input: - data: data set containing a column with date values for every 48 rows                        #
#        - time_zone: character string indicating the time zone (UTC, MEZ,...)                          #
#        - date_column: quoted character string containing the column name of the date values           # 
#                                                                                                       #
# Output: newData: data set whose first column contains date-time objects of the analysers measuring    #
#         times and the respective measurements                                                         #
#                                                                                                       #
# Use: The function "createTimeAnalysers" takes as argument the input data set "data", fragments it     #
#      in blocks, one block corresponding to one day. The date of the current block will be repeated 48 #
#      times (because a day consists of 48 half hours). Theoutput data set will be created, whose first #
#      column named "time" is a concatenation of the repeated dates and the half hours from 00:00:00 up #
#      to 23:30:00 for each block. This date-time values will than be parsed into date-time objects.    #
#      The remaining columns contain the analysers measurements.                                        #
#                                                                                                       #
# Note: This function checks simultaneously if each block contains actually 48 rows.                    #
#       If a block with more or less than 48 rows appears, a warning message will be written and the    #
#       function stops executing. An NULL" object will be returned.                                     #
#                                                                                                       #
#########################################################################################################

createTimeAnalysers<-function(data,time_zone="UTC",date_column){
  
  #Repeat date for each day
  i=1
  while(i<=nrow(data)){
    MyDate<-data[i,date_column]
    i<-i+1
    while(is.na(data[i,date_column]) & i<=nrow(data)){
      data[i,date_column]<-MyDate
      i<-i+1
    }
    
    if(i%%48==1){
      
      #Add half hours
      firstDate<-data[!is.na(data[,date_column]),date_column]
      time <- seq(parse_date_time(paste(firstDate[1],"00:00:00"),orders="dmy HMS",tz=time_zone),
                  parse_date_time(paste(firstDate[length(firstDate)],"23:30:00"),orders="dmy HMS",tz=time_zone), 
                  by=1800)
      
      #1800 seconds = 30 minutes = 1 half hour
      #we are sur that start is at 00:00:00 the first day up to 23:30:00 the last day, 
      #because we checked that each block contains 48 rows!
      d=which(colnames(data)==date_column)
      newData<-data.frame(time,data[,-d])
      return(newData)
    } 
    
    else{
      warning("Attention: nombre de demi-heures non correct!")
      break
    }
  }
}


#########################################################################################################
#                                                                                                       #
#                                   9. Creation of date-time object                                     #
#                                   ===============================                                     #
#                                                                                                       #
# Input: - data_analysers: data frame containing a first column \texttt{time} with date-time objects    #
#                          and measurements in the remaining columns                                    #
#        - data_sensors: data frame with the same specifications but with another time step in the      #
#                        first column                                                                   #
#        - time_step_analysers: integer indicating which time difference is used in "data_analysers"    # 
#        - time_step_sensors: integer indicating which time difference is used in "data_sensors"        # 
#                                                                                                       #
# Output: data_approx: data set containing interpolated measurements in terms of the smaller time steps #                                                        #
#                                                                                                       #
# Use: The function "interpolation" compares the time steps of both data sets in the input, then        #
#      performs a linear interpolation on the data set with the larger time step and returns an output  #
#      data set named "data_approx" which contains the interpolated measurements in terms of the smaller#
#      time steps.                                                                                      #
#########################################################################################################

interpolation<-function(data_analysers,data_sensors,time_step_analysers,time_step_sensors){
  
  if(time_step_analysers>time_step_sensors){
    data_interpolation=data_analysers
    data_approx=data_sensors$time
  }
  
  else{
    data_interpolation=data_sensors
    data_approx=data_analysers$time
  }
  
  interpolationX=data_approx 
  
  for(i in 2:ncol(data_interpolation)){
    approximation<-approx(data_interpolation$time,data_interpolation[,i],xout=interpolationX,method="linear")$y
    data_approx<-data.frame(data_approx,approximation)
  }
  colnames(data_approx)<-colnames(data_interpolation)
  return(data_approx)
}


#########################################################################################################
#                                                                                                       #
#                                     10. Fusion of the data sets                                       #
#                                     ===========================                                       #
#                                                                                                       #
# Input: - data_approximated: data frame containing the approximated values by interpolation with first #
#                             column "time"                                                             #
#        - data_exact: data frame containing the real measured values by with first column "time"       #                                                                 #
#        - colNames: character string containing the chosen column names for the output data            #
#                                                                                                       #
# Output: data_fusion: data set whose first column contains the date-time objects, the remaining columns#
#         contain the interpolated measurements from "data_approximated" and then the exact measurements#                                                        #
#                                                                                                       #
# Use: The function "dataFusion" merges the data set containing the exact measured values with the      #
#      approximated data set.                                                                           #
#                                                                                                       #
# Note: The column names of the output data set can be chosen for the convenience and clearness for     #
#       the subsequent work.                                                                            #
#########################################################################################################


dataFusion<-function(data_approximated,data_exact,colNames){
  
  data_fusion<-data_approximated$time
  data_fusion<-data.frame(data_fusion, data_approximated[,-1],data_exact[,-1])
  colnames(data_fusion)=colNames
  
  return(data_fusion)
}


#########################################################################################################
#                                                                                                       #
#                                        11. Reduce time steps                                          #
#                                        =====================                                          #
#                                                                                                       #
# Input: - data: data frame containing in the first column date-time objects                            #
#        - step: strictly positive integer specifying the desired number of minutes between each        #
#                measurement                                                                            #
#                                                                                                       #
# Output: smallData: data set whose first column contains the date-time objects whose differences are   #
#                    given by the argument "step"                                                       #
#                                                                                                       #
# Use: The number of minutes in the argument "step" is added to the first time value in the data set,   #
#      and the smallest time value greater than this sum is defined as the next measuring time.         #
#      This process continues until the end of the data set.                                            #   
#                                                                                                       #
#########################################################################################################

timeReduction<-function(data, step){
  
  smallData <- data[1,]
  nextDate <- data[1,1] + minutes(step)
  for(i in 2:nrow(data)){
    if(data[i,1] >= nextDate){
      smallData<-rbind(smallData,data[i,])
      nextDate <- data[i,1] + minutes(step)
    }
  }
  return(smallData)
}


#########################################################################################################
#                                                                                                       #
#                                      12. Delete observations                                          #
#                                      =======================                                          #
#                                                                                                       #
# Input: - data: data frame with first column containing a time series                                  #
#        - start_rows: vector containing quoted character strings specifying the starting time of the   #
#                      intervals to delete in the format year-month-day hour:minute:second.             #
#        - end_rows: vector containing quoted character strings specifying the ending time of the       #
#                      intervals to delete in the format year-month-day hour:minute:second.             #
#                                                                                                       #
# Output: newData: data set whose first column contains the date-time objects whose differences are     #
#                    given by the argument "step"                                                       #
#                                                                                                       #
# Use: The function "deleteRows" returns the data set "newData" containing only the rows of the input   #
#      data set, whose measuring time is not between two respective starting and ending times of        #
#      intervals.                                                                                       #   
#                                                                                                       #
#########################################################################################################

deleteRows<-function(data,start_rows,end_rows){
  newData=data
  cat('The data set contains',nrow(data),'lines.\n\n')
  
  start_rows<-parse_date_time(start_rows,orders="ymd HMS")
  end_rows<-parse_date_time(end_rows,orders="ymd HMS")
  
  if(length(start_rows)!=length(end_rows)){
    warning("The number of starting and end times must be identical!\n
            The data file will not be changed!")
    return(newData)
  }
  
  else{
    interval <- 1
    toDrop <- NULL
    for(i in 1:nrow(newData)){
      
      if(newData$time[i] >= start_rows[interval]){
        
        if(newData$time[i] <= end_rows[interval]){
          toDrop <- c(toDrop, i)
        }
        
        if(newData$time[i] >= end_rows[interval]){
          if(interval  == length(end_rows)){
            break
          }
          
          else{
            interval <- interval + 1
          }
        }
      }
    }
    
    newData <- newData[-toDrop,]
    cat('The new data set contains',nrow(newData),'lines after deletion.\n')
    return(newData)
  }
}


