
##############################################################################################################
#                                                                                                            #
#                                   Shiny application for one data set                                       #
#                                   ==================================                                       # 
#                                                                                                            #
# Luisa Mennicken                                                                                            #
# Master thesis                                                                                              #
#                                                                                                            #
##############################################################################################################

### Necessary libraries 

library(shiny)
library(lubridate)


######################################### Preprocessing functions ############################################

### 1. Concatenation of sensors files                                    
### ================================= 

sensorsFusion<-function(files_location,save_location,date){
  
  ### Create an empty data set. If it existed before, it will be emptied.
  
  file.create(save_location)
  files<-list.files(files_location)
  
  for(i in 1:length(files)){
    mon_doc<-paste0(paste0(files_location,"/"),files[i])
    f<-read.table(mon_doc,header=TRUE)
    
    if(i==1){
      date_1<-(read.table(file=mon_doc,header=TRUE,nrows=1)[,date])
      d1<-parse_date_time(date_1,orders="dmy")
      write.table(f,save_location,append=TRUE,col.names=TRUE,row.names=FALSE,dec=",",sep=";")
    }
    
    else{
      date_i<-(read.table(file=mon_doc,header=TRUE,nrows=1)[,date])
      di<-parse_date_time(date_i,orders="dmy")
      
      ### Check if the current date has been already seen before
      
      if(di==d1){
        cat(files[i])
        warning("Repeated day!")
      }
      
      else{
        d1<-di
        write.table(f,save_location,append=TRUE,col.names=FALSE,row.names=FALSE,dec=",",sep=";")
      }
    }
  }
  
  completeData=read.csv2(save_location,header=TRUE)
  return(completeData)
}



### 2. Elimination of useless data                            
### ==============================

elimination<-function(data,var_to_keep){
  
  ### Check if the names are correct column names of the data
  notPossible<-which(!(var_to_keep %in% names(data)))
  
  if(length(notPossible)>0){
    warning("The specified columns to keep are not in the data set!")
    return(data)
  }
  
  else{
    newData<-data[ , var_to_keep]
  }
  return(newData)
}



### 3. Parsing into date-time object                                    
### ================================  

createTimeSensors<-function(data,time_zone,date_column,time_column){
  
  time<-parse_date_time(paste(data[,date_column],data[,time_column]),orders="dmy HMS",tz=time_zone)
  del=which(colnames(data)==date_column | colnames(data)==time_column)
  newData<-data.frame(time,data[,-del])
  return(newData)
}


### 4. Conductance transformation                     
### =============================

conductance<-function(data,columns){
  
  notPossible<-which(!(columns %in% names(data)))
  
  if(length(notPossible)>0){
    warning("The specified columns to inverse are not possible!")
    newData=data
  }
  else{
    inverse<-function(x){
      return(1/x)
    }
    
    newData=data
    newData[,columns]=sapply(newData[,columns],inverse)
  }
  return(newData)
}


### 5. Missing values                                        
### =================

addRowsNA<-function(data){
  
  data$Dt = c(diff(data[,1]),0) # return the time difference in seconds - last one = 0 because no following value
  units(data$Dt) <-"mins"
  data$Dt=round(data$Dt)
  
  
  if (sum(data$Dt >1) >0){  # the following code is applied if we have at least one missing value
    
    lastMesureBeforeMissing = data$time[data$Dt >1] 
    firstMesureAfterMissing = data$time[which(data$Dt>1)+1]
    
    ###  Definition of the missing intervals 
    IntervalsMissing = cbind.data.frame(round(lastMesureBeforeMissing + minutes(1),units="mins"), 
                                        round(firstMesureAfterMissing - minutes(1),units="mins")) 
    
    colnames(IntervalsMissing) = c("Start", "End")
    
    cat('Intervals where the values are missing \n')
    cat('-------------------------------------- \n')
    print(IntervalsMissing)
    cat('-------------------------------------- \n')
    
    ### Preparation of the NA lignes to add
    
    ### Time list
    MissingDataTime = NULL
    for (i in 1:nrow(IntervalsMissing)){
      cat("i=",i,'\n')
      
      MissingDataTime= as_datetime(c(MissingDataTime, 
                                     seq(IntervalsMissing$Start[i], IntervalsMissing$End[i], by ="min")))
    }
    
    cat('Number of added lines in the file \n')
    cat('--------------------------------- \n')
    cat('N =', length(MissingDataTime), '\n')
    cat('Nb Obs file input :', nrow(data), '\n')
    cat('Expected Nb Obs file output :', nrow(data) + length(MissingDataTime), '\n')
    cat('--------------------------------- \n')
    
    ### Creation of data.frame with intercalated missings
    dataSupp = data.frame(MissingDataTime)
    names(dataSupp) = 'time'
    newData = merge(x= data, y= dataSupp, by= 'time', all.x= T, all.y = T)
    
    cat('Nb Obs file output :', nrow(newData), '\n')
    cat('--------------------------------- \n')
    
    
  }    
  else{newData = data}
  
  newData<-newData[,-ncol(newData)]
  return(newData)
}


### 6. Read and save the analysers file                               
### ===================================

readAndSave<-function(file_path){ 
  
  headerAna1 <- scan(file_path, nlines = 1,sep=";",what = character())
  headerAna2<-scan(file_path, nlines = 1,sep=";",what = character(),skip=1)
  pAna<-paste0(headerAna1,headerAna2)
  
  data_ana<- read.csv2(file_path, 
                       skip = 2,col.names=pAna,header = FALSE,na.strings="",sep=";",dec=",")
  
  return(data_ana)
}


### 7. Elimination of useless data          
### ============================== 

# same function as 2. (see line 64-81)


### 8. Creation of date-time object                           
### ===============================

createTimeAnalysers<-function(data,time_zone="UTC",date_column){
  
  #Repeat date for each day
  i=1
  while(i<=nrow(data)){
    MyDate<-data[i,date_column]
    i<-i+1
    while(is.na(data[i,date_column]) & i<=nrow(data)){
      data[i,date_column]<-MyDate
      i<-i+1
    }
    
    if(i%%48==1){
      
      #Add half hours
      firstDate<-data[!is.na(data[,date_column]),date_column]
      time <- seq(parse_date_time(paste(firstDate[1],"00:00:00"),orders="dmy HMS",tz=time_zone),
                  parse_date_time(paste(firstDate[length(firstDate)],"23:30:00"),orders="dmy HMS",tz=time_zone), 
                  by=1800)
      
      #1800 seconds = 30 minutes = 1 half hour
      #we are sur that start is at 00:00:00 the first day up to 23:30:00 the last day, 
      #because we checked that each block contains 48 rows!
      d=which(colnames(data)==date_column)
      newData<-data.frame(time,data[,-d])
      return(newData)
    } 
    
    else{
      warning("Attention: nombre de demi-heures non correct!")
      break
    }
  }
}


### 9. Creation of date-time object                               
### ===============================

interpolation<-function(data_analysers,data_sensors,time_step_analysers,time_step_sensors){
  
  if(time_step_analysers>time_step_sensors){
    data_interpolation=data_analysers
    data_approx=data_sensors$time
  }
  
  else{
    data_interpolation=data_sensors
    data_approx=data_analysers$time
  }
  
  interpolationX=data_approx 
  
  for(i in 2:ncol(data_interpolation)){
    approximation<-approx(data_interpolation$time,data_interpolation[,i],xout=interpolationX,method="linear")$y
    data_approx<-data.frame(data_approx,approximation)
  }
  colnames(data_approx)<-colnames(data_interpolation)
  return(data_approx)
}


### 10. Fusion of the data sets                                   
### ===========================

dataFusion<-function(data_approximated,data_exact,colNames){
  
  data_fusion<-data_approximated$time
  data_fusion<-data.frame(data_fusion, data_approximated[,-1],data_exact[,-1])
  colnames(data_fusion)=colNames
  
  return(data_fusion)
}



### 11. Reduce time steps                                     
### =====================

timeReduction<-function(data, step){
  
  smallData <- data[1,]
  nextDate <- data[1,1] + minutes(step)
  for(i in 2:nrow(data)){
    if(data[i,1] >= nextDate){
      smallData<-rbind(smallData,data[i,])
      nextDate <- data[i,1] + minutes(step)
    }
  }
  return(smallData)
}


### 12. Delete observations                                          
### =======================

deleteRows<-function(data,start_rows,end_rows){
  newData=data
  cat('The data set contains',nrow(data),'lines.\n\n')
  
  start_rows<-parse_date_time(start_rows,orders="ymd HMS")
  end_rows<-parse_date_time(end_rows,orders="ymd HMS")
  
  if(length(start_rows)!=length(end_rows)){
    warning("The number of starting and end times must be identical!\n
            The data file will not be changed!")
    return(newData)
  }
  
  else{
    interval <- 1
    toDrop <- NULL
    for(i in 1:nrow(newData)){
      
      if(newData$time[i] >= start_rows[interval]){
        
        if(newData$time[i] <= end_rows[interval]){
          toDrop <- c(toDrop, i)
        }
        
        if(newData$time[i] >= end_rows[interval]){
          if(interval  == length(end_rows)){
            break
          }
          
          else{
            interval <- interval + 1
          }
        }
      }
    }
    
    newData <- newData[-toDrop,]
    cat('The new data set contains',nrow(newData),'lines after deletion.\n')
    return(newData)
  }
}


############################################# Shiny interface ################################################

ui <- fluidPage(
  fluidRow(
    
    h1("Graphical interface"),
    
    textInput("time_zone","Insert the used time zone for the measurements")),
  
    textOutput("instructions"),
  
    titlePanel("First data set"),
    sidebarLayout(
    sidebarPanel(
      width=2,
      
      # Sensors fusion
      tags$h4("1. Fusion of the sensors data"),
      textInput("dir","Insert the input directory of the location of the files to concatenate"),
      textInput("save_dir","Insert the save location of the new data set\n with file name and extension (.csv):"),
      textInput("date_fusion_sen","Insert the column name of the date values"),
      actionButton("sensorsFusion","Ok"),
      p("Click the button to validate your entries."),
      
      tags$hr(),
      
      # Select variables
      tags$h4("2. Elimination of useless data"),
      uiOutput("checkboxElimSen"),
      actionButton("elimSen","Ok"),
      
      tags$hr(),
      
      # Creation of date-time object
      tags$h4("3. Creation of date-time objects"),
      textInput("time_column_sen","Insert the column name of the time values"),
      actionButton("senTime","Ok"),
      
      # Select variables for conductance
      tags$hr(),
      tags$h4("4. Conductance transformation"),
      uiOutput("checkboxConductance"),
      actionButton("conductance","Ok"),
      
      tags$hr(),
      
      # Add missing values
      tags$h4("5. Missing values"),
      actionButton("rowsNA","Ok"),
      p("Click the button to add the NA values.")
    ),
    
    mainPanel(
      titlePanel("First data set"),
      tableOutput("dataSen"),
      titlePanel("Selected data"),
      DT::dataTableOutput("elimSen"),
      titlePanel("Data with date-time objects"),
      DT::dataTableOutput("date_time_file_sen"),
      titlePanel("Data after conductance transformation"),
      DT::dataTableOutput("dataConductance"),
      titlePanel("Data after adding NA rows"),
      DT::dataTableOutput("dataWithNA")
    )
  ),
  
  sidebarLayout(
  sidebarPanel(
      width=2,
      
      # Time lag and delete observations
      tags$h4("Time lag and delete observations"),
      numericInput("timeStepFusion","Insert the desired time lag in minutes of the final data set",value=30),
      textInput("start", "Enter a vector with the starts of desired time intervals to delete. 
                If you do not want to delete any observations, type 'no'."),
      textInput("end", "Enter a vector with the ends of the desired time intervals to delete.
                If you do not want to delete any observations, type 'no'."),
      actionButton("deleteRows","Ok")
    ),
    
    mainPanel(
      titlePanel("Data set with deleted observations"),
      DT::dataTableOutput("deletedRowsData")
    )
  ),
  
  # Storage and plots of the final data
  titlePanel("Plots of the final data"),
  sidebarLayout(
  sidebarPanel(
      width=2,
      textInput("save_final","Insert the save location of the final data set\n with file name and extension (.csv):"),
      actionButton("save_final_button","Save"),
      
      tags$hr(),
      
      uiOutput("plotColumn"),
      actionButton("plot_button","Plot")
   ),
    
  mainPanel(
      plotOutput("timeSeriesPlot"),
      downloadButton(outputId = "down", label = "Download the plot")
    )
  )
)



server <- function(input, output){
  
  output$instructions<-renderText(return("Follow the instructions on the left side to execute the preprocessing.\n
                                The results are shown on the right side."))
  
  observeEvent(input$sensorsFusion,{
    output$dataSen <- renderTable({
      
      data1=sensorsFusion(input$dir,input$save_dir,input$date_fusion_sen)
      
      output$checkboxElimSen <- renderUI({
        checkboxGroupInput(inputId = "select_var_sen",
                           label = "Select variables",
                           choices = names(data1),
                           selected = names(data1))
      })
      
      observeEvent(input$elimSen,{
        output$elimSen<-DT::renderDataTable({
          
          dataAfterElimination=elimination(data1,input$select_var_sen)
          
          observeEvent(input$senTime,{
            output$date_time_file_sen<-DT::renderDataTable({
              
              sensorsDateTime=createTimeSensors(dataAfterElimination,input$time_zone,date_column=input$date_fusion_sen,
                                                time_column = input$time_column_sen)
              
              output$checkboxConductance <- renderUI({
                checkboxGroupInput(inputId = "select_var_conductance",
                                   label = "Select variables for conductance transformation",
                                   choices = names(sensorsDateTime))
              })
              
              observeEvent(input$conductance,{
                output$dataConductance<-DT::renderDataTable({
                  
                  dataCond=conductance(sensorsDateTime,input$select_var_conductance)
                  
                  observeEvent(input$rowsNA,{
                    output$dataWithNA<-DT::renderDataTable({
                      
                      dataNA=addRowsNA(dataCond)
                      
                      observeEvent(input$deleteRows,{
                        output$deletedRowsData<-DT::renderDataTable({
                          
                          timeStepData=timeReduction(dataNA,input$timeStepFusion)
                          
                          if(input$start=='no' && input$end=='no') deletedRowsData=timeStepData
                          else deletedRowsData=deleteRows(timeStepData,unlist(strsplit(input$start,",")),
                                                          unlist(strsplit(input$end,",")))
                          
                          observeEvent(input$save_final_button,{
                            write.csv2(deletedRowsData,input$save_final,row.names=FALSE)})
                          
                          output$plotColumn<-renderUI({
                            selectInput("column", "Choose variable to plot:",
                                        choices=colnames(deletedRowsData)[-1])
                          })
                          
                          observeEvent(input$plot_button,{
                            output$timeSeriesPlot<-renderPlot({
                              
                              p<-function(){plot(x=deletedRowsData[,1],y=deletedRowsData[,input$column],type="l",
                                                 main=paste("Time series of",colnames(deletedRowsData[input$column])),
                                                 xlab=paste("time [",input$timeStepFusion,"] min"),
                                                 ylab=paste("concentration of", colnames(deletedRowsData[input$column])))}
                              
                              output$down <- downloadHandler(
                                filename =  function() {
                                  paste(colnames(deletedRowsData[input$column]), "pdf", sep=".")
                                },
                                content = function(file) {
                                  pdf(file)
                                  p()
                                  dev.off() 
                                } 
                              )
                              
                              return(p())
                            })
                          })
                          
                          return(deletedRowsData)
                        })
                      })
                      
                      return(dataNA)
                    })
                  })
                  
                  return(dataCond)
                })
              })
              
              return(sensorsDateTime)
            })
          })
          
          return(dataAfterElimination)
        })
      })
      
      return(head(data1))
    })
  })
}


### Shiny application

shinyApp(ui, server)



